/*******************************************************************************
 * Copyright (c) 2018-2020 Xilinx, Inc. and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * You may elect to redistribute this code under either of these licenses.
 *******************************************************************************/

/*
 * This file contains "define" statements that control agent configuration.
 * This version of configuration file uses default settings from tcf/framework/config.h
 */

#ifndef D_config
#define D_config

#include <tcf/framework/mdep.h>
#include <tcf/framework/config.h>

#endif /* D_config */
